#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#include <windows.h>

/*
#define UINT8 unsigned char
#define UINT16 unsigned short
#define UINT32 unsigned int
#define INT8 signed char
#define INT16 signed short
#define INT32 signed int
*/


typedef union
{                   /* eight general registers */
    UINT16 w[8];    /* viewed as 16 bits registers */
    UINT8  b[16];   /* or as 8 bit registers */
} necbasicregs;

typedef struct
{
	necbasicregs regs;
 	UINT16	sregs[4];

	UINT16	ip;

	INT32	SignVal;
	UINT32  AuxVal, OverVal, ZeroVal, CarryVal, ParityVal; /* 0 or non-0 valued flags */
	UINT8	TF, IF, DF, MF; 	/* 0 or 1 valued flags */	/* OB[19.07.99] added Mode Flag V30 */
	UINT32	int_vector;
	UINT32	pending_irq;
	UINT32	nmi_state;
	UINT32	irq_state;
	int     (*irq_callback)(int irqline);
} nec_Regs;

extern nec_Regs I;

#include "source/nec/nec.h"
#include "source/nec/necmodrm.h"

FILE *trace=0;
UINT8 *rom_mapped;
UINT8 *ram_mapped;

int pc_switch[2];

extern char hook_trace;
extern int trace_map;

extern FILE* fp_hook;
extern unsigned int rd_mode[2], wr_mode[2], pc_mode[2];
extern unsigned int rd_low[2],rd_high[2];
extern unsigned int wr_low[2],wr_high[2];
extern unsigned int pc_low[2],pc_high[2];

extern int trace_map;
extern void Dis_8086(char *buffer, UINT32 address);

extern UINT8 *ws_ioRam;
extern unsigned long romSize;
#define IO_ROM_BANK_BASE_SELECTOR	0xC0

void DeInitDebug()
{
  if(rom_mapped) free(rom_mapped);
  if(ram_mapped) free(ram_mapped);
	if(trace) fclose(trace);
}

void classify( int *type, int bank ) {
	if( bank>=0 && bank<=1 ) *type = 1;
	else *type = 2;
}

void Memory_8086(int pc_type, int pc_addr, FILE *fp_hook)
{
	if(pc_type==1) fprintf(fp_hook,"[%01x:04x] ", pc_addr>>16, pc_addr&0xffff);
	if(pc_type==2) fprintf(fp_hook,"[%06x] ", pc_addr);
	if(pc_type==3) fprintf(fp_hook,"[i/o $%02x] ", pc_addr);
}

void Trace_8086(int address)
{
	int bank,offset;
	int type;
	char buffer[1024];

  if( !trace ) {
    trace=fopen("trace.log", "w");

		rom_mapped = (UINT8*) malloc(0x800000);
    memset(rom_mapped,0,0x800000);
 
		ram_mapped = (UINT8*) malloc(0x20000);
    memset(ram_mapped,0,0x20000);
  }

	bank = ( ( I.sregs[CS]<<4)+I.ip ) >> 16;
	offset = ( ( I.sregs[CS]<<4)+I.ip ) & 0xffff;

	switch( bank ) {
	case 0:
	case 1:
		break;
	case 2:
	case 3:
		address = offset+((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+bank]&((romSize>>16)-1))<<16);
	default: 
		int romBank=(256-(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(bank&0xf)));
		address = offset+romSize-(romBank<<16);
	}

	classify( &type, bank );

	if( type==1 && ram_mapped[address] ) return;
	if( type==2 && rom_mapped[address] ) return;

	if( type==1 ) ram_mapped[address] = 1;
	if( type==2 ) rom_mapped[address] = 1;

	memset( buffer, ' ', 1024 );

	Memory_8086( type, address, trace );
	fprintf( trace, "%01x:%04x ", bank & 0xf, offset );

	Dis_8086( buffer, 0 );
	fprintf( trace, "%-43s", buffer );

	fprintf( trace, " ax=%04x bx=%04x cx=%04x dx=%04x || bp=%04x sp=%04x si=%04x di=%04x || cs=%04x ds=%04x ss=%04x es=%04x || 2=%02x 3=%02x || 4=%02x 5=%02x 6=%02x 7=%02x 8=%02x 9=%02x a=%02x b=%02x c=%02x d=%02x e=%02x f=%02x",
		I.regs.w[AW], I.regs.w[BW], I.regs.w[CW], I.regs.w[DW], 
		I.regs.w[BP], I.regs.w[SP], I.regs.w[IX], I.regs.w[IY], 
		I.sregs[CS], I.sregs[DS], I.sregs[SS], I.sregs[ES],

		ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+2]&((romSize>>16)-1),
		ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+3]&((romSize>>16)-1),

		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x4 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x5 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x6 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x7 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x8 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x9 & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xa & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xb & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xc & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xd & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xe & 0xf))&((romSize>>16)-1),
		(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xf & 0xf))&((romSize>>16)-1)
		);

	fprintf( trace, "\n" );
}

void Hook_Read(char mode, unsigned int byte, unsigned int addr)
{
	int type;
	UINT32 pc_addr,pc_type;

	pc_addr = (I.sregs[CS]<<4)+I.ip;
	if( mode != 2 ) classify(&type,addr>>16);
	else type = 3;

	if( type == 2 ) {
		int bank = ( ( I.sregs[CS]<<4)+I.ip ) >> 16;
		int offset = ( ( I.sregs[CS]<<4)+I.ip ) & 0xffff;

		switch( bank ) {
		case 0:
		case 1:
			break;
		case 2:
		case 3:
			addr = offset+((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+bank]&((romSize>>16)-1))<<16);
		default: 
			int romBank=(256-(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(bank&0xf)));
			addr = offset+romSize-(romBank<<16);
		}
	}

	for( int lcv = 0; lcv < 2; lcv++ ) {
		// contiguous RAM
		if(type==1 && rd_mode[lcv]==0 && rd_low[lcv]<=addr && addr<=rd_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] R08 = %02x [R-%05X]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
		// contiguous ROM
		else if(type==2 && rd_mode[lcv]==1 && rd_low[lcv]<=addr && addr<=rd_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] R08 = %02x [%06X]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
		// i/o
		else if(type==3 && rd_mode[lcv]==2 && rd_low[lcv]<=addr && addr<=rd_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] R08 = %02x [i/o $%02x]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
	}
}

void Hook_Write(int mode, unsigned int byte, unsigned int addr)
{
	int type;
	UINT32 pc_addr,pc_type;

	pc_addr = (I.sregs[CS]<<4)+I.ip;
	if( mode != 2 ) classify(&type,addr>>16);
	else type = 3;

	if( type == 2 ) {
		int bank = ( ( I.sregs[CS]<<4)+I.ip ) >> 16;
		int offset = ( ( I.sregs[CS]<<4)+I.ip ) & 0xffff;

		switch( bank ) {
		case 0:
		case 1:
			break;
		case 2:
		case 3:
			addr = offset+((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+bank]&((romSize>>16)-1))<<16);
		default: 
			int romBank=(256-(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(bank&0xf)));
			addr = offset+romSize-(romBank<<16);
		}
	}

	for( int lcv = 0; lcv < 2; lcv++ ) {
		// contiguous RAM
		if(type==1 && wr_mode[lcv]==0 && wr_low[lcv]<=addr && addr<=wr_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] W08 = %02x [R-%05X]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
		// contiguous ROM
		else if(type==2 && wr_mode[lcv]==1 && wr_low[lcv]<=addr && addr<=wr_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] W08 = %02x [%06X]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
		// i/o
		else if(type==3 && wr_mode[lcv]==2 && wr_low[lcv]<=addr && addr<=wr_high[lcv])
		{
			fprintf(fp_hook,"[%01x:%04x] W08 = %02x [i/o $%02x]\n",
				pc_addr>>16, pc_addr&0xffff, byte, addr);
		}
	}
}

void Hook_PC()
{
	int type;
	UINT32 pc_addr,pc_type;
	char buffer[4092];

	int bank = ( ( I.sregs[CS]<<4)+I.ip ) >> 16;
	int offset = ( ( I.sregs[CS]<<4)+I.ip ) & 0xffff;

	pc_addr = (I.sregs[CS]<<4)+I.ip;
	classify(&type,pc_addr>>16);

	if( type == 2 ) {
		switch( bank ) {
		case 0:
		case 1:
			break;
		case 2:
		case 3:
			pc_addr = offset+((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+bank]&((romSize>>16)-1))<<16);
		default: 
			int romBank=(256-(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(bank&0xf)));
			pc_addr = offset+romSize-(romBank<<16);
		}
	}

	for( int lcv = 0; lcv < 2; lcv++ ) {
		// contiguous RAM
		if(type==1 && pc_mode[lcv]==0 && pc_low[lcv]==pc_addr) {
			pc_switch[lcv] = 1;
		}
		// contiguous ROM
		else if(type==2 && pc_mode[lcv]==1 && pc_low[lcv]==pc_addr) {
			pc_switch[lcv] = 1;
		}

		if( !pc_switch[lcv] ) continue;

		memset( buffer, ' ', 1024 );

		Memory_8086( type, pc_addr, fp_hook );
		fprintf( fp_hook, "%01x:%04x ", bank & 0xf, offset );

		Dis_8086( buffer, 0 );
		fprintf( fp_hook, "%-43s", buffer );

		fprintf( fp_hook, " ax=%04x bx=%04x cx=%04x dx=%04x || bp=%04x sp=%04x si=%04x di=%04x || cs=%04x ds=%04x ss=%04x es=%04x || 2=%02x 3=%02x || 4=%02x 5=%02x 6=%02x 7=%02x 8=%02x 9=%02x a=%02x b=%02x c=%02x d=%02x e=%02x f=%02x",
			I.regs.w[AW], I.regs.w[BW], I.regs.w[CW], I.regs.w[DW], 
			I.regs.w[BP], I.regs.w[SP], I.regs.w[IX], I.regs.w[IY], 
			I.sregs[CS], I.sregs[DS], I.sregs[SS], I.sregs[ES],

			ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+2]&((romSize>>16)-1),
			ws_ioRam[IO_ROM_BANK_BASE_SELECTOR+3]&((romSize>>16)-1),

			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x4 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x5 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x6 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x7 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x8 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0x9 & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xa & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xb & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xc & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xd & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xe & 0xf))&((romSize>>16)-1),
			(((ws_ioRam[IO_ROM_BANK_BASE_SELECTOR]&0xf)<<4)|(0xf & 0xf))&((romSize>>16)-1)
			);

		fprintf( fp_hook, "\n" );

		// contiguous RAM
		if(type==1 && pc_mode[lcv]==0 && pc_high[lcv]==pc_addr) {
			pc_switch[lcv] = 0;
		}
		// contiguous ROM
		else if(type==2 && pc_mode[lcv]==1 && pc_high[lcv]==pc_addr) {
			pc_switch[lcv] = 0;
		}
	}
}
